/* aBolo
 * Copyright (C) 2000 Josh Lauricha
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "aBolo.h"
#include <sys/timeb.h>
#include "string.h"
#include "unistd.h"


/* I don't like linked-lists... so why do I have so 4? */
DIRTY_LIST     dirty;
EXPLOSION_LIST explodes;
MINE_LIST      mines;
WATER_LIST     water;

MAP		map;
BITMAP	 *tiles;           /* BITMAP of all the map tiles    */
BITMAP	 *buffer;          /* Buffer to draw/blit too        */
BITMAP	 *player_tank;     /* BITMAP sprite of the gray tank */
BITMAP	 *boat;	           /* BITMAP sprite of the boat	     */
BITMAP	 *explosion[EXPLODE_FRAMES];

PALETTE	 tiles_palette;	   /* Colour PALETTE for the tiles   */

int		 frames;           /* 'Frame' counter                */
int		 lite_bevel;
int		 norm_bevel;       /* palette entries for the interface colours.*/
int		 dark_bevel;
int		 SCREEN_TL_X = 80;  /* These designate where the TopLeft  */
int		 SCREEN_TL_Y = 150; /* of the map screen is	              */

TANK		players[MAX_TANKS];
TANK	   *tank;           /* Pointer to the local player's tank
                               in the above array                 */
int fps_timer  = 0;
int move_timer = 0;
int keys_timer = 0;
int boom_timer = 0;
int mine_timer = 0;
int water_timer= 0;


void add_dirty_rect(int X, int Y, int W, int H) {
	struct DIRTY_RECT *node;

	node = (DIRTY_RECT *) malloc(sizeof(DIRTY_RECT));
	node->x = X;
	node->y = Y;
	node->w = W;
	node->h = H;
	node->next = NULL;
	if(dirty.tail) {
		dirty.tail->next = node;
	} else {
		dirty.head = node;
	}
	dirty.tail = node;
}

void destroy_dirty() {
	struct DIRTY_RECT *node = dirty.head, *dest = NULL;
	while(node) {
		dest = node;
		node = node->next;
		free(dest);
	}
	dirty.head = NULL;
	dirty.tail = NULL;
}

void add_explosion(int x, int y) {
	struct EXPLOSION *node;

	node = (EXPLOSION *) malloc(sizeof(EXPLOSION));
	node->x = x;
	node->y = y;
	node->frame = 0;
	node->next = NULL;
	if(explodes.head) {
		explodes.tail->next = node;
	} else {
		explodes.head = node;
	}
	explodes.tail = node;
}

void destroy_explodes() {
	struct EXPLOSION *node = explodes.head, *dest = NULL;
	while(node) {
		dest = node;
		node = node->next;
		free(dest);
	}
	explodes.head = NULL;
	explodes.tail = NULL;
}

void add_mine(int x, int y) {
	struct MINE *node;

	node = (MINE *) malloc(sizeof(MINE));
	node->x = x;
	node->y = y;
	node->next = NULL;
	if(mines.new_head) {
		mines.new_tail->next = node;
	} else {
		mines.new_head = node;
	}
	mines.new_tail = node;
}

void destroy_mines() {
	struct MINE *node, *dest = NULL;

	if(mines.head) mines.tail->next = mines.new_head;
	else mines.head = mines.new_head;

	node = mines.head;
	while(node) {
		dest = node;
		node = node->next;
		free(dest);
	}
	mines.head = NULL;
	mines.tail = NULL;
	mines.new_head = NULL;
	mines.new_tail = NULL;
}

void add_water(int x, int y) {
	struct WATER *node;

	node = (WATER *) malloc(sizeof(WATER));
	node->x = x;
	node->y = y;
	node->next = NULL;
	if(water.new_head) {
		water.new_tail->next = node;
	} else {
		water.new_head = node;
	}
	water.new_tail = node;
}

void destroy_water() {
	struct WATER *node, *dest = NULL;

	if(water.head) water.tail->next = water.new_head;
	else water.head = water.new_head;

	node = water.head;
	while(node) {
		dest = node;
		node = node->next;
		free(dest);
	}
	water.head = NULL;
	water.tail = NULL;
	water.new_head = NULL;
	water.new_tail = NULL;
}


void draw_screen() {
/* This blits the 'dirty' parts of the buffer to the screen
 * much faster than drawing the entire screen */
	struct DIRTY_RECT *node = dirty.head, *dest = NULL;

	acquire_screen();
   
	while(node) {
		blit(buffer, screen, node->x, node->y,
			 node->x, node->y, node->w,
			 node->h);
	dest = node;
	node = node->next;
	free(dest);
	}
	dirty.head = NULL;
	dirty.tail = NULL;

	release_screen();
}

void draw_bevel(int x, int y, int w, int h, unsigned char pressed) {
/* This function draws the beveled rectangles used for almost all bolo
 * displays. */

	int points[8];
	int c1, c2;

	if(pressed) {
		c1 = dark_bevel;
		c2 = lite_bevel;
	} else {
		c1 = lite_bevel;
		c2 = dark_bevel;
	}

	points[0] = x - 4;	 points[1] = y - 4;
	points[2] = x + w + 4; points[3] = y - 4;
	points[4] = x + w;	 points[5] = y;
	points[6] = x;		 points[7] = y;

	polygon(buffer, 4, points, c1);

	points[0] = x - 4;	 points[1] = y - 4;
	points[2] = x - 4;	 points[3] = y + h + 4;
	points[4] = x;		 points[5] = y + h;
	points[6] = x;		 points[7] = y;

	polygon(buffer, 4, points, c1);

	points[0] = x + w + 4; points[1] = y - 4;
	points[2] = x + w + 4; points[3] = y + h + 4;
	points[4] = x + w;	 points[5] = y + h;
	points[6] = x + w;	 points[7] = y;

	polygon(buffer, 4, points, c2);

	points[0] = x - 4;	 points[1] = y + h + 4;
	points[2] = x + w + 4; points[3] = y + h + 4;
	points[4] = x + w;	 points[5] = y + h;
	points[6] = x;		 points[7] = y + h;

	polygon(buffer, 4, points, c2);

	rectfill(buffer, x, y, x + w, y + h, 0);
	
	add_dirty_rect( x-4, y-4, w+8, h+8);
}

void draw_buttons() {
/* Draws the buttons to the screen. */
	static unsigned char pressed[5] = {0,0,1,0,0};

	draw_bevel(BUTTON1_SCREEN_X, BUTTON1_SCREEN_Y, BUTTON_W, BUTTON_H, pressed[0]);
	draw_bevel(BUTTON2_SCREEN_X, BUTTON2_SCREEN_Y, BUTTON_W, BUTTON_H, pressed[1]);
	draw_bevel(BUTTON3_SCREEN_X, BUTTON3_SCREEN_Y, BUTTON_W, BUTTON_H, pressed[2]);
	draw_bevel(BUTTON4_SCREEN_X, BUTTON4_SCREEN_Y, BUTTON_W, BUTTON_H, pressed[3]);
	draw_bevel(BUTTON5_SCREEN_X, BUTTON5_SCREEN_Y, BUTTON_W, BUTTON_H, pressed[4]);


	 blit(tiles,			 buffer,
		 BUTTON1_TILES_X,   BUTTON1_TILES_Y,
		 BUTTON1_SCREEN_X,  BUTTON1_SCREEN_Y,
		 BUTTON_W,		  BUTTON_H);

	blit(tiles,			 buffer,
		 BUTTON2_TILES_X,   BUTTON2_TILES_Y,
		 BUTTON2_SCREEN_X,  BUTTON2_SCREEN_Y,
		 BUTTON_W,		  BUTTON_H);

	blit(tiles,			 buffer,
		 BUTTON3_TILES_X,   BUTTON3_TILES_Y,
		 BUTTON3_SCREEN_X,  BUTTON3_SCREEN_Y,
		 BUTTON_W,		  BUTTON_H);

	blit(tiles,			 buffer,
		 BUTTON4_TILES_X,   BUTTON4_TILES_Y,
		 BUTTON4_SCREEN_X,  BUTTON4_SCREEN_Y,
		 BUTTON_W,		  BUTTON_H);

	blit(tiles,			 buffer,
		 BUTTON4_TILES_X,   BUTTON4_TILES_Y,
		 BUTTON4_SCREEN_X,  BUTTON4_SCREEN_Y,
		 BUTTON_W,		  BUTTON_H);

	blit(tiles,			 buffer,
		 BUTTON5_TILES_X,   BUTTON5_TILES_Y,
		 BUTTON5_SCREEN_X,  BUTTON5_SCREEN_Y,
		 BUTTON_W,		  BUTTON_H);
}


void quit_program() {
/* Uninits Allegro and exits the program */
	int x=0;
	printf("Game closed\n");
	remove_keyboard();
	
	destroy_bitmap(tiles);
	destroy_bitmap(buffer);
	destroy_bitmap(player_tank);
	destroy_bitmap(boat);
	destroy_dirty();
	destroy_explodes();
	destroy_mines();
	destroy_water();

	for( ; x<EXPLODE_FRAMES; x++) {
		destroy_bitmap(explosion[x]);
	}
		
	allegro_exit();
	exit(0);
}

void tank_shoot() {
	if(tank->ammo) tank->ammo--;
}

void tank_start() {
		/* Set the screen position: */
	if(map.s != 0) {
		/* Get a random start point: */
		struct timeb time_bomb;
		int i = 0;
		#ifdef WIN32
		 ftime(&time_bomb);
		#else
		 (void)ftime(&time_bomb);
		#endif
		srand(time_bomb.millitm + time_bomb.time);
		i = (int)(((double)map.s-1)*rand()/RAND_MAX+1.0);
				
		SCREEN_TL_X = map.start[i].x-7;
		SCREEN_TL_Y = map.start[i].y-7;

		tank->x	  = map.start[i].x;
		tank->y	  = map.start[i].y;
		tank->dir	 = map.start[i].dir;
		tank->tx	 = 128;
		tank->ty	 = 128;
		tank->armour = MAX_TANK_ARMOUR;
		tank->ammo   = MAX_TANK_AMMO;
		tank->mines  = MAX_TANK_MINES;
		tank->trees  = MAX_TANK_TREES;
		tank->speed  = 0;
		tank->onBoat = 1;
	} else {
		fprintf(stderr, "ERROR: Invalid map: No start points\n");
		exit(1);
	}
}

void tank_dead(TANK *myTank) {
	memset( (void *)myTank, 0, sizeof(TANK));
	tank_start(myTank);
}

void shot_damage(int x, int y, int shots) {
	int i=0;
	
	if(shots > 3) {
		fprintf(stderr, "Invalid shot damage: %i\n", shots);
		return;
	}
	
	for(;i<MAX_TANKS; i++) {
		if(players[i].x == x && players[i].y == y) {
			players[i].armour -= shots;
			if(players[i].armour <= (unsigned char)0) {
				tank_dead(&players[i]);
			}
			break;
		}
	}
	// CHECK PILLS
	i = ((map.data[x][y] & 0xf00)>>8) - shots;
	if(i <= 0) {
		static BYTE downers[11] = {
			8,   3,      3,    3,      6,     7,      3,      2,     6,   1,    10 };
		/*BUILD, RIVER, SWAMP, CRATER, ROAD, FOREST, RUBBLE, GRASS, DMG, Boat, Deep */
		map.data[x][y] = downers[map.data[x][y] & 0xf];
		if(map.data[x][y] == 0x3) {
			BYTE a, watermask = 0;
			a = map.data[x][y-1] & 0xf;
			if( IS_WATER(a)   ) watermask |= 1;

			a = map.data[x+1][y] & 0xf;
			if( IS_WATER(a)   ) watermask |= 2;

			a = map.data[x][y+1] & 0xf;
			if( IS_WATER(a)   ) watermask |= 4;
	
			a = map.data[x-1][y] & 0xf;
			if( IS_WATER(a)   ) watermask |= 8;
			
			if(watermask) add_water(x,y);
		}		
	} else {
		map.data[x][y] = (map.data[x][y] ^ 0xf00) + ((i & 0xf)<<8);
	}
		
}

void tank_set_mine() {
	if(tank->mines && !(map.data[tank->x][tank->y] & 0x10)
	   && (map.data[tank->x][tank->y]&0xf) != 10
   	   && (map.data[tank->x][tank->y]&0xf) != 1 ) {
		map.data[tank->x][tank->y] |= 0x30;
	}	
}

void check_keys() {
/* Called 10 times a second
 * Checks for keyboard input and takes actions based on that */
	int k;
	
	while(keypressed()) {
		k = readkey() & 0xFF;
		if(k == ' ') tank_shoot();
		if(k == 'q') quit_program();
	}

	/* Rotate tank left */
	if(key[KEY_A]) tank->dir = (tank->dir +1) %16;

	/* Accellerate */
	if(key[KEY_W]) if(tank->speed < 240) tank->speed += 16;
	
	/* Rotate tank right */
	if(key[KEY_D]) tank->dir = (tank->dir-1)%16;

	/* Deccellerate */
	if(key[KEY_S]) if(tank->speed) tank->speed -= 8;

	if(key[KEY_B]) {
		add_explosion(tank->x, tank->y);
	}

	if(key[KEY_TAB]) tank_set_mine();

	if(key[KEY_N]) tank_start();

	if(key[KEY_SPACE]) tank_shoot();
	
	if(key[KEY_Q]) quit_program();
}


void check_mines() {
	int x, y;
	x = tank->x - MINE_DIST;
	y = tank->y - MINE_DIST;

	if(HAS_MINE(tank->x,tank->y)) {
		add_mine(tank->x, tank->y);
	}
		
	for(; x <= tank->x + MINE_DIST; x++) {
		for(; y <= tank->y + MINE_DIST; y++) {
			if(HAS_MINE(x,y)) {
				map.data[x][y] |= 0x20;
			}
		}
	}
}

void move_stuff() {
/*   BLD  RIVER SWAMP CRTER RD   FREST RBBLE GRASS HALFB BOAT DEEPS */
	static int max_speeds[] = 
	{0,   48,   48,   48,   240, 96,   48,   192,  0,	48,  1 };
	int dir = tank->dir;
	int terrain = map.data[tank->x][tank->y] & 0xf;
	BYTE old_x = tank->x, old_y = tank->y, newTile=0;

	if(tank->speed > max_speeds[terrain] && terrain != 10 && terrain != 1)
		tank->speed -= 8;
	else if((terrain == 10 || terrain == 1) 
			&& tank->onBoat && tank->speed > max_speeds[4]) tank->speed -= 8;
	else if(!tank->onBoat && terrain == 1 
			&& tank->speed > max_speeds[1]) tank->speed -= 8;
	
	dir = dir + 4;
	if(dir > 15) dir = dir % 16;
	if(12 != dir) 
		 tank->ty = tank->ty
				  + ((tank->speed>>4)*cos((double)(dir<<4)*M_PI/128));
	tank->tx = tank->tx + ((tank->speed>>4)*sin((double)(dir<<4)*M_PI/128));
	
	if(tank->ty > 255) {
		if(CAN_DRIVE_ON(map.data[tank->x][tank->y+1])) {
			int old_ty = tank->ty;
			tank->ty &= 0xFF;
			old_ty -= tank->ty;
			tank->y += old_ty>>8;
		} else tank->ty = 254; /* If they can't drive on it push
								* the tank up to the edge of it. */
	} else if (tank->ty < 0) {
		if(CAN_DRIVE_ON(map.data[tank->x][tank->y-1])) {	
			int old_ty = tank->ty;
			tank->ty &= 0xFF;
			old_ty -= tank->ty;
			tank->y += old_ty>>8;
		} else tank->ty = 1;
	}

	if(tank->tx > 255) {
		if(CAN_DRIVE_ON(map.data[tank->x+1][tank->y])) {
			int old_tx = tank->tx;
			tank->tx &= 0xFF;
			old_tx -= tank->tx;
			tank->x += old_tx>>8;
			newTile = TRUE;
		} else tank->tx = 254;
	} else if (tank->tx < 0) {
		if(CAN_DRIVE_ON(map.data[tank->x-1][tank->y])) {
			int old_tx = tank->tx;
			tank->tx &= 0xFF;
			old_tx -= tank->tx;
			tank->x += old_tx>>8;
			newTile = TRUE;
		} else tank->tx = 1;
	}
				
	/* Primitive AutoScroll, will eventually move to center on the nearest
	 * enemy tank/pill */

	if(tank->x - 4 < SCREEN_TL_X || tank->x + 4 > SCREEN_TL_X+15)
		SCREEN_TL_X = tank->x - 8;
	if(tank->y - 4 < SCREEN_TL_Y || tank->y + 4 > SCREEN_TL_Y+15)
		SCREEN_TL_Y = tank->y - 8;

	/* Handle boat [dis]embarking */
	if(tank->onBoat && (map.data[tank->x][tank->y] & 0xf) != 10 
	   && (map.data[tank->x][tank->y] & 0xf) != 1) {
		tank->onBoat = 0;
		map.data[old_x][old_y] = 9;
	}
	if((map.data[tank->x][tank->y] & 0xf) == 9) {
		map.data[tank->x][tank->y] = 1;
		tank->onBoat = 1;
	}

	terrain = map.data[tank->x][tank->y] & 0xf;
	if(!(tank->onBoat) && 10 == terrain) {
		tank_dead(tank);
	}


	if(newTile) {
		check_mines();
	
	}

	/* Make draw_map() actually draw the thing */
	MAP_CHANGED = TRUE;
}


void draw_stats() {
	static BYTE old_armour, old_ammo, old_mines, old_trees;
	if(old_ammo != tank->ammo) {
		int x1, y1, x2, y2;
		x1 = AMMOICON_SCREEN_X+AMMOICON_W/2-1;
		x2 = AMMOICON_SCREEN_X+AMMOICON_W/2+1;
		y2 = AMMOICON_SCREEN_Y;
		y1 = y2 - (((double)tank->ammo/MAX_TANK_AMMO)*ICON_BAR_H);
		rectfill(buffer, x1, y2 - ICON_BAR_H, x2, y2, 0);
		if(tank->ammo) rectfill(buffer, x1, y1, x2, y2, GOOD_COLOUR);
		add_dirty_rect(x1, y2 - ICON_BAR_H, x2, y2); 
		old_ammo = tank->ammo;
	}
	
	if(old_armour != tank->armour) {
		int x1, y1, x2, y2;
		x1 = SHIELDICON_SCREEN_X+SHIELDICON_W/2-1;
		x2 = SHIELDICON_SCREEN_X+SHIELDICON_W/2+1;
		y2 = SHIELDICON_SCREEN_Y;
		y1 = y2 - (((double)tank->armour/MAX_TANK_ARMOUR)*ICON_BAR_H);
		rectfill(buffer, x1, y2 - ICON_BAR_H, x2, y2, 0);
		if(tank->armour) rectfill(buffer, x1, y1, x2, y2, GOOD_COLOUR);
		add_dirty_rect(x1, y2 - ICON_BAR_H, x2, y2); 
		old_armour = tank->armour;	
	}
	
	if(old_mines != tank->mines) {
		int x1, y1, x2, y2;
		x1 = MINEICON_SCREEN_X+MINEICON_W/2-1;
		x2 = MINEICON_SCREEN_X+MINEICON_W/2+1;
		y2 = MINEICON_SCREEN_Y;
		y1 = y2 - (((double)tank->mines/MAX_TANK_MINES)*ICON_BAR_H);
		rectfill(buffer, x1, y2 - ICON_BAR_H, x2, y2, 0);
		if(tank->mines) rectfill(buffer, x1, y1, x2, y2, GOOD_COLOUR);
		add_dirty_rect(x1, y2 - ICON_BAR_H, x2, y2); 
		old_mines = tank->mines;
	}
	
	if(old_trees != tank->trees) { 
		int x1, y1, x2, y2;
		x1 = TREEICON_SCREEN_X+TREEICON_W/2-1;
		x2 = TREEICON_SCREEN_X+TREEICON_W/2+1;
		y2 = TREEICON_SCREEN_Y;
		y1 = y2 - (((double)tank->trees/MAX_TANK_TREES)*ICON_BAR_H);
		rectfill(buffer, x1, y2 - ICON_BAR_H, x2, y2, 0);
		if(tank->trees) rectfill(buffer, x1, y1, x2, y2, GOOD_COLOUR);
		add_dirty_rect(x1, y2 - ICON_BAR_H, x2, y2); 
		old_trees = tank->trees;
	}
}

void first_draw() {
/* This clears the screen and redraws everything. */
	clear_to_color(buffer, norm_bevel);
	clear_to_color(screen, norm_bevel);

	/* Draw the screen */
	draw_buttons();

	draw_bevel(MESSAGEBOX_X,MESSAGEBOX_Y,
			   MESSAGEBOX_W,MESSAGEBOX_H,1);

	draw_bevel(MAP_X,	   MAP_Y,
			   MAP_W,	   MAP_H,	   1);

	draw_bevel(WORLDSTATS_X,WORLDSTATS_Y,
			   WORLDSTATS_W,WORLDSTATS_H,1);

	draw_bevel(LOCALSTATS_X,LOCALSTATS_Y,
			   LOCALSTATS_W,LOCALSTATS_H,1);

	draw_map();
	
	/* This section is in the dirty_list from the bevel
	 * so I don't need to readd it */
	blit(tiles, buffer,
		 AMMOICON_TILES_X,  AMMOICON_TILES_Y,
		 AMMOICON_SCREEN_X, AMMOICON_SCREEN_Y,
		 AMMOICON_W,		AMMOICON_H);

	blit(tiles, buffer,
		 MINEICON_TILES_X,  MINEICON_TILES_Y,
		 MINEICON_SCREEN_X, MINEICON_SCREEN_Y,
		 MINEICON_W,		MINEICON_H);

	blit(tiles, buffer,
		 SHIELDICON_TILES_X,  SHIELDICON_TILES_Y,
		 SHIELDICON_SCREEN_X, SHIELDICON_SCREEN_Y,
		 SHIELDICON_W,		SHIELDICON_H);

	blit(tiles, buffer,
		 TREEICON_TILES_X,  TREEICON_TILES_Y,
		 TREEICON_SCREEN_X, TREEICON_SCREEN_Y,
		 TREEICON_W,		TREEICON_H);

	draw_screen();
}

void generate_explosions() {
   #define BMP_SIZE 64
   BITMAP *bmp = create_bitmap(BMP_SIZE, BMP_SIZE);
   unsigned char *p;
   int c, c2;
   int x, y;
   int xx, yy;
   int color;

   #define HOTSPOTS  128

   struct HOTSPOT {
	  int x, y;
	  int xc, yc;
   } hot[HOTSPOTS];

   for (c=0; c<HOTSPOTS; c++) {
	  hot[c].x = hot[c].y = (BMP_SIZE/2)<<16;
	  hot[c].xc = (rand()&0xFFFF)-0x7FFF;
	  hot[c].yc = (rand()&0xFFFF)-0x7FFF;
   }

   for (c=0; c<EXPLODE_FRAMES; c++) {
	  clear(bmp);

	  color = ((c<EXPLODE_FRAMES/4) ? c*4 : (80-c)) >> 2;

	  for (c2=0; c2<HOTSPOTS; c2++) {
	 for (x=-6; x<=6; x++) {
		for (y=-6; y<=6; y++) {
		   xx = (hot[c2].x>>16) + x;
		   yy = (hot[c2].y>>16) + y;
		   if ((xx>0) && (yy>0) && (xx<BMP_SIZE) && (yy<BMP_SIZE)) {
		  p = bmp->line[yy] + xx;
		  *p += (color >> ((ABS(x)+ABS(y))/3));
		  if (*p > 63)
			 *p = 63;
		   }
		}
	 }
	 hot[c2].x += hot[c2].xc;
	 hot[c2].y += hot[c2].yc;
	  }

	  for (x=0; x<BMP_SIZE; x++) {
	 for (y=0; y<BMP_SIZE; y++) {
		c2 = bmp->line[y][x];
		if (c2 < 8)
		   bmp->line[y][x] = 0;
		else
		   bmp->line[y][x] = 16+c2/4;
	 }
	  }

	  stretch_blit(bmp, explosion[c], 1, 1, BMP_SIZE, BMP_SIZE,
				   1, 1, EXPLODE_SIZE, EXPLODE_SIZE);
   }
   destroy_bitmap(bmp);
}

void draw_tanks() {
	if(tank->onBoat) {
		rotate_sprite(buffer, boat,
			((tank->x-SCREEN_TL_X)<<4)+MAP_X + (tank->tx>>4)-8,
			((tank->y-SCREEN_TL_Y)<<4)+MAP_Y + (tank->ty>>4)-8,
			itofix((16-tank->dir+4)<<4)); 
	}

	rotate_sprite(buffer, player_tank,
		((tank->x-SCREEN_TL_X)<<4)+MAP_X + (tank->tx>>4)-8,
		((tank->y-SCREEN_TL_Y)<<4)+MAP_Y + (tank->ty>>4)-8,
		itofix((16-tank->dir+4)<<4));
	add_dirty_rect(
		((tank->x-SCREEN_TL_X)<<4)+MAP_X + (tank->tx>>4)-8,
		((tank->y-SCREEN_TL_Y)<<4)+MAP_Y + (tank->ty>>4)-8,
		16,16);
}

void draw_explosions() {
	static struct EXPLOSION *node;
	node = explodes.head;
	
	while(node) {
		if(node->x >= SCREEN_TL_X && node->x < (SCREEN_TL_X + 15)
		   && node->y >= SCREEN_TL_Y && node->y < (SCREEN_TL_Y + 15)) {
			masked_blit(explosion[node->frame], buffer, 1, 1,
					((node->x-SCREEN_TL_X)<<4)+MAP_X,
					((node->y-SCREEN_TL_Y)<<4)+MAP_Y, 
					EXPLODE_SIZE, EXPLODE_SIZE);
			add_dirty_rect(
					((node->x-SCREEN_TL_X)<<4)+MAP_X,
					((node->y-SCREEN_TL_Y)<<4)+MAP_Y, 
					EXPLODE_SIZE, EXPLODE_SIZE);

		}
		node = node->next;
	}
}

void move_explosions() {
	static struct EXPLOSION *node, *dest, *temp;
	node = explodes.head;
	
	while(node) {
		node->frame++;
		if(node->frame >= EXPLODE_FRAMES) {
			dest = node;
			temp = node;
			node = node->next;
			if(temp != explodes.head) {
				temp = explodes.head;
				while(temp && temp->next != dest)
					temp = temp->next;
				if(temp)
					temp->next = dest->next;
				else 
					fprintf(stderr, "Error deleting explosion node: no prev not head\n");
			} else {
				explodes.head = temp->next;
			}
			if(explodes.tail == dest) {
				if(temp)
					explodes.tail = temp->next;
				else
					explodes.tail = NULL;
			}
			free(dest);
		}

		if(node) node = node->next;
	}
	MAP_CHANGED = TRUE;
}

void move_mines() {
	MINE *node;
	int x,y, i=0;
	
	if(!mines.new_head) return;
	
	if(mines.head) {
		mines.tail->next = mines.new_head;
		mines.tail = mines.new_tail;
	} else {
		mines.head = mines.new_head;
		mines.tail = mines.new_tail;
	}
	mines.new_head = NULL;
	mines.new_tail = NULL;
	
	while(mines.head) {
		x = mines.head->x;
		y = mines.head->y;
		node = mines.head;
		mines.head = mines.head->next;
		free(node);
		map.data[x][y] &= 0xf;
		
		add_explosion(x,y);
		map.data[x][y] = 0x3;
		shot_damage(x,y,3);
		printf("Mine: %i\n", i++);

		if(HAS_MINE(x-1,y)) {
			add_mine(x-1,y);
		}
		if(HAS_MINE(x+1,y)) {
			add_mine(x+1,y);
		}
		if(HAS_MINE(x,y-1)) {
			add_mine(x,y-1);
		}
		if(HAS_MINE(x,y+1)) {
			add_mine(x,y+1);
		}
	}
}

void move_water() {
	WATER *node;
	int x,y, i=0;
	
	if(!water.new_head) return;
	
	if(water.head) {
		water.tail->next = water.new_head;
		water.tail = water.new_tail;
	} else {
		water.head = water.new_head;
		water.tail = water.new_tail;
	}
	water.new_head = NULL;
	water.new_tail = NULL;
	
	while(water.head) {
		x = water.head->x;
		y = water.head->y;
		node = water.head;
		water.head = water.head->next;
		free(node);
		
		map.data[x][y] = 0x1;
		printf("Water: %i\n", i++);

		if(IS_CRATER(x-1,y)) {
			add_water(x-1,y);
		}
		if(IS_CRATER(x+1,y)) {
			add_water(x+1,y);
		}
		if(IS_CRATER(x,y-1)) {
			add_water(x,y-1);
		}
		if(IS_CRATER(x,y+1)) {
			add_water(x,y+1);
		}
	}
}

void init() {
/* Initialize the general game */
	int i;
	for(i = 0; i < MAX_TANKS; i++) {
		map.tanks[i].num = i;
	}
	tank = map.tanks;

	map_read("default.map");
	tank_start();
	/* Setup allegro */
	allegro_init();
	install_keyboard();
	install_timer();


	for(i=0; i < EXPLODE_FRAMES; i++)
		explosion[i] = create_bitmap(EXPLODE_SIZE, EXPLODE_SIZE);

	/* this section is needed becaus the directX drivers don't like
	 * aBolo for some reason
	 */

	#ifndef WIN32
	 set_gfx_mode(GFX_AUTODETECT, WIDTH, HEIGHT, 0, 0);
	#else
	 set_gfx_mode(GFX_GDI, WIDTH, HEIGHT, 0, 0);
	#endif

	/* Read & setup bitmaps */
	buffer = create_bitmap(WIDTH,HEIGHT);
	tiles  = load_bmp("tiles.bmp"  , tiles_palette);

	set_palette(tiles_palette);
	
	player_tank = create_bitmap(16, 16);
	boat		= create_bitmap(16, 16);

	blit(tiles, player_tank, 1, 306, 0, 0, 16, 16); 
	blit(tiles,		boat, 17, 306, 0, 0, 16, 16); 

	/* Setup global variables */

	lite_bevel = getpixel(tiles, 230, 262);
	norm_bevel = getpixel(tiles, 235, 262);
	dark_bevel = getpixel(tiles, 240, 262);

	generate_explosions();
	explodes.head = explodes.tail = NULL;
	dirty.head = dirty.tail = NULL;
	first_draw();
}

void fps() {
	printf("FPS: %i\n", frames);
	frames = 0;
}

void bolo_loop() {
/* This is the main bolo game loop. */

	/* install the timers */
	fps_timer  = install_int(&fps,           1000); /* 1x  a second */
	move_timer = install_int(&move_stuff,      20); /* 50x a second */
	keys_timer = install_int(&check_keys,     100); /* 10x a second */
	mine_timer = install_int(&move_mines,     250); /* 10x a second */
	water_timer = install_int(&move_water,    250); /* 10x a second */
	boom_timer = install_int(&move_explosions, 20); /* 50x a second */
	
	frames = 0;

	do {
		if(MAP_CHANGED) {
			/* Update the buffer: */
			draw_map();
			draw_tanks();
			draw_explosions();
			draw_stats();
			MAP_CHANGED = FALSE;
		}
		
		/* Update the screen: */
		draw_screen();
		frames++; 
	} while( 1 );
}

int main(int argc, char *argv[]) {

int counti;
char datapath[256];

for (counti=strlen(argv[0]) ; argv[0][counti]!='/' ; counti--);
strncpy(datapath, argv[0], counti);
chdir(datapath);

	init();
	bolo_loop();

	/* Just incase 1 becomes false... */
	quit_program();
	
	return(0);
}

END_OF_MAIN();
