/* aBolo
 * Copyright (C) 2000 Josh Lauricha
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "aBolo.h"

BYTE  nibble_flag;
BYTE *nibble_data;
BYTE  MAP_CHANGED = TRUE;

void map_clear() {
/* Clears the map to all deep sea, with sea mines at the borders */
	int x, y;
	/* Main ocean block */
	for(x = 10; x < MAPSIZE - 10; x++)
		for(y = 10; y < MAPSIZE - 10; y++)
			map.data[x][y] = 0xa;

	/* Top */
	for(x = 0; x < MAPSIZE; x++)
		for(y = 0; y < 10; y++)
			map.data[x][y] = 0x3a;

	/* Left */
	for(x = 0; x < 10; x++)
		for(y = 10; y < MAPSIZE; y++)
			map.data[x][y] = 0x3a;

	/* Right */
	for(x = MAPSIZE - 10; x < MAPSIZE; x++)
		for(y = 10; y < MAPSIZE - 10; y++)
			map.data[x][y] = 0x3a;

	/* Bottom */
	for(x = 0; x < MAPSIZE; x++)
		for(y = MAPSIZE - 10; y < MAPSIZE; y++)
			map.data[x][y] = 0x3a;
}

BYTE map_get_tile(BYTE x, BYTE y) {
/* Finds the x,y tile-coords of the tile (in the tiles bitmap) and
 * returns a BYTE with the first nibble being x, second being y.
 * Most of this was liberated from http://bolo.mersh.com and
 * bastardized to work with my code. Ok, so some small changes were
 * made, like actually using quick_returns[]. The comments are basically
 * all his from this point down.
 */

/*   BLD  RIVER SWAMP CRTER RD   FREST RBBLE GRASS HALFB BOAT  DEEPS */
	static BYTE quick_returns[] =
	{0x00,0x00, 0x00, 0x00, 0x00,0x00, 0x86, 0x06, 0x08, 0x00, 0x00};

	static BYTE adjval[] =
	{0x01,0x02, 0x04, 0x08, 0x10,0x20, 0x40, 0x00, 0x01, 0x02, 0x80};

	static BYTE possadj[] =
		{
		0x41,   /* BUILDING	   rubble or any other building       */
		0x9A,   /* RIVER       deepsea, road, crater or river     */
		0x86,   /* SWAMP       deepsea, swamp or river            */
		0x8A,   /* CRATER      deepsea, crater or river           */
		0x10,   /* ROAD        road only                          */
		0x20,   /* FOREST      forest only                        */
		0x00,   /* RUBBLE       -                                 */
		0x00,   /* GRASS        -                                 */
		0x00,   /* HALFBUILDING -                                 */
		0x8B,   /* BOAT        deepsea, crater, river or building */
		};

	static BYTE watermask, adj4, adj8, a, t;
	watermask = adj4 = adj8 = a = 0;
	t = map.data[x][y] & 0xf;


	/* simple tiles with no adjacency rules: */
	if(quick_returns[t]) return(quick_returns[t]);

	/* Check the 4-way adj: */
	a = map.data[x][y-1] & 0xf;
	if( IS_WATER(a)   ) watermask |= 1;
	if( adjval[a] & possadj[t] ) adj4	  |= 1;


	a = map.data[x+1][y] & 0xf;
	if( IS_WATER(a)   ) watermask |= 2;
	if( adjval[a] & possadj[t] ) adj4	  |= 2;

	a = map.data[x][y+1] & 0xf;
	if( IS_WATER(a)   ) watermask |= 4;
	if( adjval[a] & possadj[t] ) adj4	  |= 4;


	a = map.data[x-1][y] & 0xf;
	if( IS_WATER(a)   ) watermask |= 8;
	if( adjval[a] & possadj[t] ) adj4	  |= 8;

	/* all except ROAD and BUILDING have 4-connected adjacency rules */
	if (t != 4 && t != 0) return(adj4<<4 | t);

	/* Screwy stuff: */
	/* ROAD and BRIDGE can have 8-connected adjacency rules */
	if (t == 4 && watermask) {
		static BYTE bridgechar[16] =
			{ 0x00, 0x40, 0x80, 0xC0, 0x10, 0xEB, 0x90, 0xEB,
			  0x20, 0x60, 0xFB, 0xFB, 0x30, 0xEB, 0xFB, 0xDB };

		if (bridgechar[watermask] & 0x0F) return(bridgechar[watermask]);
		if ((bridgechar[watermask] & (adj4<<4)) == bridgechar[watermask])
			return(bridgechar[watermask] | 0x0B);
	}

	if ((adj4 & 0x9) == 0x9 && (adjval[ map.data[x-1][y-1]&0x0f ]
		& possadj[t]))
		adj8 |= 8;
	if ((adj4 & 0xC) == 0xC && (adjval[ map.data[x-1][y+1]&0x0f ]
		& possadj[t]))
		adj8 |= 4;
	if ((adj4 & 0x6) == 0x6 && (adjval[ map.data[x+1][y+1]&0x0f ]
		& possadj[t]))
		adj8 |= 2;
	if ((adj4 & 0x3) == 0x3 && (adjval[ map.data[x+1][y-1]&0x0f ]
		& possadj[t]))
		adj8 |= 1;

	if (!adj8)
		return(adj4<<4 | t); /* no diagonal adjacency: normal character */

	if (t == 4) return(0x06 | (adj4<<4)); /* lineless road ('car park') */

	/* OK, clever stuff here; Building with diagonal adjacency */

	if (adj4 == 0x3) return(0x07);  /* 4 orientations of L shape */
	if (adj4 == 0x6) return(0x47);
	if (adj4 == 0x9) return(0x87);
	if (adj4 == 0xC) return(0xC7);

	if (adj4 == 0xF) return(0x8 | adj8<<4);   /* X building with flat bits */

	if (adj4 == 0x7) return(0x07 + (adj8<<4));		/* adj8 = 1,2,3 */
	if (adj4 == 0xE) return(0x47 + (adj8<<3));	  /* adj8 = 2,4,6 */
	if (adj4 == 0xD) return(0x87 + (adj8<<2));	  /* adj8 = 4,8,C */
	/* for this case adj8 = 8,1,9 so if we just shift we would lose a bit */

	adj8 = adj8<<1;
	if(adj8 & 0x10) adj8 = (adj8 - 0x10) + 0x1;
	if (adj4 == 0xB) return(0xC7 + (adj8<<4) );	  /* adj8 = 8, 10, 18 */

	/* If, for some odd reason, control got to this point, return deep sea */
	return(0x0a);
}

void map_set_tile(BYTE x, BYTE y, BYTE nub) {
/* This takes the mapfile numbers and changes it to the map.data format */
	if(nub < 10) map.data[x][y] = nub;
	else {
		/* its got a mine */
		nub ^= 0x8; /* remove the last bit */
		nub |= 0x30; /* add the mine bit */
		map.data[x][y] = nub;
	}
}

void draw_map() {
/* Complete redraw (onto buffer) of the map area */
	int x, y, i;
	BYTE type;

	for(x = 0; x < 15; x++) {
		for(y = 0; y < 15; y++) {
			type = map_get_tile(x + SCREEN_TL_X, y + SCREEN_TL_Y);
			blit(tiles, buffer, 
				 ((type & 0xf0))+1, ((type & 0x0f)<<4)+1,
				 (x<<4) + MAP_X,	   (y<<4) + MAP_Y,
				 16, 16);

				/* Flag check: */
			if(map.data[x + SCREEN_TL_X][y+SCREEN_TL_Y] & 0x20) {
				/* Visible Mine:*/
				masked_blit(tiles, buffer, (4<<4)+1, (6<<4)+1,
					 (x<<4) + MAP_X, (y<<4) + MAP_Y,
					  16, 16);
			}

			if(map.data[x + SCREEN_TL_X][y + SCREEN_TL_Y] & 0x40) {
				/* Pill: */
				i = 0;
				while(map.p > i && ( map.pills[i].x != x+SCREEN_TL_X
					  || map.pills[i].y != y+SCREEN_TL_Y) ) i++;
					if(map.pills[i].owner == tank->num) type = 0x0F;
				else type = 0x0C;

				type |= (map.pills[i].armour&0xf)<<4;

				masked_blit(tiles, buffer,
					(type&0xf0)+1, ((type&0x0f)<<4)+1,
					(x<<4) + MAP_X,
					(y<<4) + MAP_Y,
					16, 16);
			}

			if(map.data[x + SCREEN_TL_X][y + SCREEN_TL_Y] & 0x80) {
				/* Base: */
				i = 0;
				while(map.b > i && ( map.bases[i].x != x+SCREEN_TL_X
					  || map.bases[i].y != y+SCREEN_TL_Y) ) i++;

				if(map.bases[i].owner == tank->num) type = 0x5B;
				else if(map.bases[i].owner == 0xFF) type = 0x0B;
				else type = 0x6B;

				blit(tiles, buffer,
					(type&0xf0)+1, ((type&0x0f)<<4)+1,
					(x<<4) + MAP_X,
					(y<<4) + MAP_Y,
					16, 16);
			}
		}
	}
	add_dirty_rect(MAP_X, MAP_Y, MAP_X+256, MAP_Y+256);
}

void map_read(char *filename) {
	char *info = "BMAPBOLO";
	FILE *file;
	BYTE x=0, c=0,y=0, i=0, datalen=0, 
		 startx=0, endx=0, nib=0, nub=0,
		 /*length=0,*/ /*nib_flag = 1,*/ buffer[0xff];

	map_clear();
	
	if( !(file = fopen(filename, "rb")) ) {
		fprintf(stderr, "Couldn't open map!\n");
		exit(1);
	} /* open 'binary' for read with error check */

	for(x = 0; x < 8; x++) {
		y = fread(&c,1,1,file);
		if(c != info[x]) {
			fprintf(stderr, "Not a valid map!\n");
			exit(1);
		}
	}


	/* I should really changes these...
	 * I'm lazy so I just copy the map file's bytes to memory starting
	 * at the first element in the array. if the compiler chooses to to 
	 * buffer the struct and changes the size (by making a 7byte struct
	 * an even 8) then the data will be offset and wont work... */

	if(!(fread(&map, 1, 4, file)) ) {
		fprintf(stderr, "Error reading map preamble\n");
		exit(1);
	}

	if( !(fread(&map.pills, 5, map.p, file)) ) {
		fprintf(stderr, "Error grabbing pill data\n");
		exit(1);
	}
	
	if( !fread(&map.bases, 6, map.b, file) ) {
		printf("Error grabbing base data\n");
		exit(1);
	}

	if( !fread(&map.start, map.s, 3, file) ) {
		printf("Error grabbing start points\n");
		exit(1);
	}


	/* Now read the actual map */

	while( 1 ) {
		/* Read the run header */
		if(!fread(&datalen, 1, 1, file) ) {
			printf("Error reading header\n");
			exit(1);
		}

		if(!fread(&buffer, 1, datalen-1, file) ) {
			printf("Couldn't load run\n");
			exit(1);
		}

		if(datalen == 4 && 0xff == buffer[0]
				&& 0xff == buffer[1] 
				&& 0xff == buffer[2]) {
				break;
		}

		nibble_data = buffer;
		y = *nibble_data++;
		startx = x= *nibble_data++;
		endx = *nibble_data++;
		nibble_flag = 0;

		while( x < endx ) {
			nib = get_nibble();
			if(nib < 8)
				do
					map_set_tile(x++, y, get_nibble());
				while(nib-- >0);
			else if(nib < 16) {
				nub = get_nibble();
				do
					map_set_tile(x++, y, nub);
				while(nib-- >7);
			} else {
				printf("NIB ERR: %i\n", nib);
				exit(1);
			}
		}
	}

	for(i = 0; i < map.b; i++) {
		map.data[map.bases[i].x][map.bases[i].y] |= 0x80;
		map.data[map.bases[i].x][map.bases[i].y] ^= 0x30;
	}
	for(i = 0; i < map.p; i++) {
		map.data[map.pills[i].x][map.pills[i].y] |= 0x40;
		map.data[map.bases[i].x][map.bases[i].y] ^= 0x30;
	}



	fclose(file);
}
